# RedHat-specific operations
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import urlgrabber.grabber as grabber

from VMLinux import VMLinux
from exceptions import *
import mount
import msg
import util

class VMRedHat(VMLinux):
    def _kernel(self, base):
        return '%s/images/xen/vmlinuz' % (base,)
    def _initrd(self, base):
        return '%s/images/xen/initrd.img' % (base,)

#    def _getHostAndPathFromNFS(self, url):
        # RedHat NFS URLs are of the form:
        # nfs:host:/path/to/install/source
        # SUSE's are of the form:
        # nfs://host/path/to/install/source

    def getInstallImagesFromUrl(self, url, progress=lambda f:None):
        """Retrieves a Xen-enabled RedHat kernel and initrd from the URL, and
        copies them to temporary files.  Returns the (kernel, initrd) tuple of
        filenames."""
        kfd = ifd = None
        try:
            try:
                if url.startswith('http://') or url.startswith('ftp://'):
                    kfd = grabber.urlopen(self._kernel(url))
                    ifd = grabber.urlopen(self._initrd(url))
                elif url.startswith('nfs:'):
                    mnt = mount.TemporaryMount(url[4:])
                    mountpoint = mnt.mount(mnt)
                    kfd = open(self._kernel(mountpoint), 'r')
                    ifd = open(self._initrd(mountpoint), 'r')
                else:
                    raise InstSrcError(err=InstSrcError.E_UNSUPPORTED, details=url)
                kfn = util.copy_to_temp(kfd, prefix='vmlinuz.', progress=lambda f:progress(f*0.5))
                try:
                    ifn = util.copy_to_temp(ifd, prefix='initrd.', progress=lambda f:progress(0.5+f*0.5))
                except:
                    os.unlink(kfn)
                    raise
            except IOError, e:
                raise RuntimeError(msg.invalid_url + ': ' + str(e))
        finally:
            if kfd: kfd.close()
            if ifd: ifd.close()
        return (kfn, ifn)

    def getInstallImagesFromPath(self, path, progress=lambda f:None):
        """Retrieves a Xen-enabled RedHat kernel and initrd from the base path, and
        copies them to temporary files.  Returns the (kernel, initrd) tuple of
        filenames."""
        kfd = ifd = None
        try:
            kfd = open(self._kernel(path), 'r')
            ifd = open(self._initrd(path), 'r')
            kfn = util.copy_to_temp(kfd, prefix='vmlinuz.', progress=lambda f:progress(f*0.5))
            try:
                ifn = util.copy_to_temp(ifd, prefix='initrd.', progress=lambda f:progress(0.5+f*0.5))
            except:
                os.unlink(kfn)
                raise
        finally:
            if kfd: kfd.close()
            if ifd: ifd.close()
        return (kfn, ifn)

    def getInstallArgs(self, options):
        """Return RedHat's kernel arguments for installs."""
        if options.full_virt:
            return ''
        args = VMLinux.getInstallArgs(self, options)
        if type(options.source) is str:
            args += ' method=%s' % options.source
        if options.os_settings:
            args += ' ks=%s' % options.os_settings
        return args

    def preprocessOptions(self, options):
        """Pulls settings out of extra_args."""
        if not options.extra_args:
            return
        args = options.extra_args.split()
        extra_args = ''
        for arg in args:
            if arg.startswith('method='):
                if not options.source:
                    src = arg[7:]
                    if src.startswith('hd://'):
                        src = 'dev:/' + src[5:]
                    options.source = src
            elif arg.startswith('ks=') and not options.os_settings:
                options.os_settings = arg[3:]
            else:
                extra_args += arg + ' '
        options.extra_args = extra_args

