# Probes capabilities of the platform
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os

from exceptions import *

def cpu_flags():
    f = open('/proc/cpuinfo')
    lines = f.readlines()
    f.close()
    for line in lines:
        if not line.startswith('flags'):
            continue
        flags = line[:-1].split(':', 1)[1]
        flst = flags.split(' ')
        return flst
    return []

def cpu_processors():
    f = open('/proc/cpuinfo')
    lines = f.readlines()
    f.close()
    num_processors = 0
    for line in lines:
        if not line.startswith('processor'):
            continue
        num_processors += 1
    if num_processors == 0:
        num_processors = 1
    return num_processors

def vcpu_processors(max_vcpus):
    cpus = cpu_processors();
    if ( cpus > max_vcpus ):
        cpus = max_vcpus
    return cpus

_arch = None
def arch():
    global _arch
    if not _arch:
        a = os.uname()[4]
        if a in ('i386', 'i586', 'i686'):
            _arch = 'i586'
        else:
            _arch = a
    return _arch

def compatible_arches():
    if is_ia64_capable():
        arches = [['ia64', 'xen']]
    elif is_x86_64_capable():
        arches = [['i386', 'xenpae'], ['i586', 'xenpae'], ['x86_64', 'xen']]
        # x86-64 can do either, but prefer to match dom0.
        if arch() == 'x86_64':
            arches.reverse()
    else:
        if is_pae_capable():
            arches = [['i386', 'xenpae'], ['i586', 'xenpae']]
        else:
            arches = [['i386', 'xen'], ['i586', 'xen']]
    return arches

def xen_arch():
    if not is_xen():
        raise XenError(XenError.E_NO_XEN)
    if is_ia64_capable():
        return 'ia64'
    if is_x86_64_capable():
        return 'x86_64'
    return 'i586'
    
_caps = []
def xen_caps():
    global _caps
    if not _caps:
        if os.path.exists('/sys/hypervisor/properties/capabilities'):
            _caps = open('/sys/hypervisor/properties/capabilities').read().split()
    return _caps

def is_ia64_capable():
    return 'xen-3.0-ia64' in xen_caps()

def is_x86_64_capable():
    return 'xen-3.0-x86_64' in xen_caps()

def is_pae_capable():
    return 'xen-3.0-x86_32p' in xen_caps()

def is_hvm_capable():
    caps = xen_caps()
    return ('hvm-3.0-x86_32' in caps or
            'hvm-3.0-x86_32p' in caps or
            'hvm-3.0-x86_64' in caps)

def is_blktap_capable():
    return os.path.exists('/dev/xen/blktap0')

def is_xen():
    return os.path.isdir('/proc/xen')

def is_dom0():
    # FIXME
    return os.path.isdir('/proc/xen')

def is_domU():
    return is_xen() and not is_dom0()
