# FIXME:  dbus calls time out.  Unknown yet if it's permissions or what.
use_dbus = False

if use_dbus:
    import dbus
import gtk
import gobject
import gtk.glade

import vminstall.msg
import vminstall.log
import vminstall.paths
import vmdisks
from tree_view import *
from chooser import *

class DisksView(PyTreeView):
    def __init__(self, disks, renumber=True):
        self.renumber = renumber
        self._pyobj_index = 4
        self.model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING,
                                   gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
        self.view = gtk.TreeView(self.model)
        column = gtk.TreeViewColumn(vminstall.msg.name, gtk.CellRendererText(), text=0)
        self.view.append_column(column)
        column = gtk.TreeViewColumn(vminstall.msg.type, gtk.CellRendererText(), text=1)
        self.view.append_column(column)
        column = gtk.TreeViewColumn(vminstall.msg.source, gtk.CellRendererText(), text=2)
        self.view.append_column(column)
        self.view.set_expander_column(column)
        column = gtk.TreeViewColumn(vminstall.msg.size_gb, gtk.CellRendererText(), text=3)
        self.view.append_column(column)
        self.set(disks)

    def _populate_row(self, iter, disk):
        vminstall.log.debug("Adding row to disk table: '%s'" % disk)
        if self.renumber:
            disk.vdev = self.model.get_path(iter)[0]
        number = disk.get_vdev_index()
        vdev_type = disk.get_vdev_type_str()
        blocks = disk.get_virtual_size()
        if blocks is not None:
            size = '%0.1f' % (blocks/(2.0*1024*1024), )
        else:
            size = '?'
        self.model.set(iter, 0, str(number+1), 1, vdev_type, 2, disk.pdev, 3, size, 4, disk)

    def get(self):
        disks = PyTreeView.get(self)
        disks.sort(lambda d1,d2: cmp(d1.vdev, d2.vdev))
        return disks

class DisksWidget:
    def __init__(self, allow_reorder=True, remove_fn=lambda disk: True,
                 edit_fn=None,
                 add_harddisk_fn=None, add_cdrom_fn=None):
        """
        remove_fn - takes a disk that user wants to remove from VM.  Return True
                    if was removed.  Set to None instead of fn to disable remove button.
        """
        self.max_disks = 0
        self.glade = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', 'disks-window')
        self.topwin = self.glade.get_widget('disks-window')
        self.topwin.hide()
        self.disks_widget = self.glade.get_widget('disks')
        self.disks_widget.get_parent().remove(self.disks_widget)
        self.topwin.destroy()
        
        # Right now, set renumber based on allow_reorder; maybe later expose
        # a separate setting
        self.renumber = allow_reorder
        
        s = {}

        self.support_reorder = allow_reorder
        if allow_reorder:
            s['on_button_disk_up_clicked'] = self._do_disk_up
            s['on_button_disk_down_clicked'] = self._do_disk_down
        else:
            child = self.glade.get_widget('reorder-disk-alignment')
            child.get_parent().remove(child)

        self.support_remove = (remove_fn is not None)
        self.remove_fn = remove_fn
        if self.support_remove:
            s['on_button_delete_disk_clicked'] = self._do_delete_disk
        else:
            child = self.glade.get_widget('button-delete-disk')
            child.get_parent().remove(child)

        self.support_edit = (edit_fn is not None)
        if edit_fn:
            s['on_button_edit_disk_clicked'] = edit_fn
        else:
            child = self.glade.get_widget('button-edit-disk')
            child.get_parent().remove(child)

        self.support_add_harddisk = (add_harddisk_fn is not None)
        if add_harddisk_fn:
            s['on_button_new_harddisk_clicked'] = add_harddisk_fn
        else:
            child = self.glade.get_widget('button-new-harddisk')
            child.get_parent().remove(child)

        self.support_add_cdrom = (add_cdrom_fn is not None)
        if add_cdrom_fn:
            s['on_button_new_cdrom_clicked'] = add_cdrom_fn
        else:
            child = self.glade.get_widget('button-new-cdrom')
            child.get_parent().remove(child)
        self.glade.signal_autoconnect(s)
        
        self.disks_view = DisksView([], self.renumber)
        self.glade.get_widget('disks-scrolled-window').get_child().add(self.disks_view.view)
        self.disks_view.view.show_all()

    def get_widget(self):
        return self.disks_widget
        
    def set_disks(self, disks):
        vminstall.log.debug("Re-loading disks table with '%s'" % map(str, disks))
        self.disks_view.set(disks)
        self.refresh()
        
    def set_max_disks(self, max_disks):
        self.max_disks = max_disks
        self.refresh()
        
    def refresh(self):
        self.glade.get_widget('disks-scrolled-window').get_child().grab_focus()
        num_disks = self.disks_view.rows()
        if self.support_reorder:
            for button in ('button-disk-up', 'button-disk-down'):
                self.glade.get_widget(button).set_sensitive(num_disks > 1)
        if self.support_edit:
            self.glade.get_widget('button-edit-disk').set_sensitive(num_disks > 0)
        if self.support_remove:
            self.glade.get_widget('button-delete-disk').set_sensitive(num_disks > 0)
        if self.support_add_cdrom:
            self.glade.get_widget('button-new-cdrom').set_sensitive(num_disks < self.max_disks)
        if self.support_add_harddisk:
            self.glade.get_widget('button-new-harddisk').set_sensitive(num_disks < self.max_disks)

    def get_disks(self):
        disks = self.disks_view.get()
        vminstall.log.debug("Retrieved disk list from disk table: '%s'" % map(str, disks))
        return disks

    def get_selected(self):
        return self.disks_view.get_selected()

    def set_index(self, index, new_disk):
        vminstall.log.debug("Setting index %d to %s" % (index, str(new_disk)))
        self.disks_view.update(index, new_disk)
    
    def append(self, disk):
        self.disks_view.append(disk)
        self.refresh()
        
    def _do_disk_up(self, button):
        for row in self.disks_view.move_selected_up():
            disk = self.disks_view.get_row(row)
            self.disks_view.update(row, disk)
        return True

    def _do_disk_down(self, button):
        for row in self.disks_view.move_selected_down():
            disk = self.disks_view.get_row(row)
            self.disks_view.update(row, disk)
        return True

    def _do_delete_disk(self, button):
        (disk, row) = self.disks_view.get_selected()
        if disk is None:
            return True
        row = self.disks_view.delete_selected()
        assert(row is not None)
        if self.renumber:
            for i in xrange(row, self.disks_view.rows()):
                disk = self.disks_view.get_row(i)
                disk.vdev = i
                self.disks_view.update(i, disk)
        if self.remove_fn is not None:
            try:
                self.remove_fn(disk)
            except:
                pass
        self.refresh()
        return True


class DiskEditWidget:
    def __init__(self):
        self.disk = None
        self.glade = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', 'disk-edit-window')
        self.topwin = self.glade.get_widget('disk-edit-window')
        self.topwin.hide()
        self.disk_edit_widget = self.glade.get_widget('disk-edit')
        self.disk_edit_widget.get_parent().remove(self.disk_edit_widget)
        self.topwin.destroy()
        
        self._default_cdrom = None
        self.disk_folder = vminstall.paths.imagedir

        prompt = vminstall.msg.prompt
        bold_prompt = vminstall.msg.bold_prompt
        small = vminstall.msg.small

        def set_text(name, widget=None, text=None, modifier=None, markup=False):
            if widget is None:
                widget = self.glade.get_widget(name)
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        def set_radio_text(name, text=None, modifier=None, markup=False):
            widget = self.glade.get_widget(name).get_children()[0]
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        set_radio_text('disk-ro', text=vminstall.msg.read_only)
        set_radio_text('disk-sparse', text=vminstall.msg.create_sparse)
        set_text('disk_help')
        set_text('disk_location', text=vminstall.msg.source, modifier=prompt)
        set_text('disk-name-prompt', text=vminstall.msg.name, modifier=prompt)
        set_text('disk_size', text=vminstall.msg.size_gb, modifier=prompt)
        set_text('virtual-disk-type', text=vminstall.msg.type, modifier=prompt)

        self.glade.signal_autoconnect({
            'on_disk_location_changed': self.disk_location_changed,
            'on_disk_gb_editing_done': self.disk_gb_editing_done,
            'on_disk_sparse_toggled': self.disk_sparse_toggled,
            'on_disk_ro_toggled': self.disk_ro_toggled,
            'on_disk_browse_clicked': self.browse_existing_disk,
            'on_disk_new_clicked': self.browse_new_disk,
        })

        # set up the list for the cd-path widget
        cd_list = self.glade.get_widget('disk-location')
        # Fields are raw device path, volume label, flag indicating
        # whether volume is present or not, and HAL path
        cd_model = gtk.ListStore(str, str, bool, str)
        cd_list.set_model(cd_model)
        cd_list.set_text_column(0)
        if not use_dbus:
            self.bus = None
            self.hal_iface = None
            self.populate_opt_media(cd_model)
        else:
            try:
                vminstall.log.debug('Getting connection to dbus.SystemBus')
                self.bus = dbus.SystemBus()
                vminstall.log.debug('Getting HAL service')
                hal_object = self.bus.get_object('org.freedesktop.Hal', '/org/freedesktop/Hal/Manager')
                self.hal_iface = dbus.Interface(hal_object, 'org.freedesktop.Hal.Manager')
                vminstall.log.debug('Enumerating CDs')
                self.populate_opt_media(cd_model)
            except Exception, e:
                vminstall.log.error("Unable to connect to HAL to list cdrom volumes: '%s'", e)
                self.bus = None
                self.hal_iface = None
                msg = gtk.MessageDialog(parent=None, flags=0, type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_OK,
                                        message_format="Unable to enumerate CDROM volumes")
                msg.format_secondary_text("Cannot connect to HAL device manager")
                msg.run()
                msg.destroy()

    def default_cdrom(self):
        return self._default_cdrom or '/dev/cdrom'

    def get_widget(self):
        return self.disk_edit_widget
        
    def set(self, disk):
        self.disk = disk
        self.refresh()
        
    def get(self):
        self.disk_location_changed(update=False)
        self.disk_gb_editing_done()
        self.disk_sparse_toggled()
        self.disk_ro_toggled()
        self.disk.validate()
        return self.disk

    def disk_location_changed(self, widget=None, update=True):
        if not widget:
            widget = self.glade.get_widget('disk-location')
        text = widget.get_active_text()
        if not text:
            return
        if self.disk.is_template() and text in (self.disk.pdev, self.disk.get_path()):
            # If it's not explicitly set, and they are setting to the default, leave it implicit.
            pass
        else:
            try:
                self.disk.pdev = text
                vminstall.log.debug("Disk location changed to '%s'" % text)
            except:
                pass
        if update:
            self.refresh(canonicalize=False)

    def disk_gb_editing_done(self, widget=None):
        if not widget:
            widget = self.glade.get_widget('disk-gb')
        gb = widget.get_value()
        blocks = long(gb*1024*1024*2)
        self.disk.set_virtual_size(blocks)

    def disk_sparse_toggled(self, widget=None):
        if not widget:
            widget = self.glade.get_widget('disk-sparse')
        self.disk.args['sparse'] = widget.get_active()

    def disk_ro_toggled(self, widget=None):
        if not widget:
            widget = self.glade.get_widget('disk-ro')
        self.disk.ro = widget.get_active()

    def refresh(self, canonicalize=True):
        button = self.glade.get_widget('disk-new-button')
        if (self.disk.vdevtype == vmdisks.Disk.DEVICE_DISK):
            button.show()
        else:
            button.hide()

        exists = self.disk.exists()
        self.glade.get_widget('disk-name').set_text(str(self.disk.get_vdev_index()+1))
        vdev_index = map(lambda row: row[0], vmdisks.Disk.translations).index(self.disk._vdevType)
        vdev_type = map(lambda row: row[1], vmdisks.Disk.translations)[vdev_index]
        self.glade.get_widget('vdev-type').set_text(vdev_type)
        if canonicalize:
            self.glade.get_widget('disk-location').get_child().set_text(self.disk.pdev)
        widget = self.glade.get_widget('disk-gb')
        blocks = self.disk.get_virtual_size()
        if blocks is not None:
            gb = blocks / (2.0*1024*1024)
        else:
            gb = 4.0
        if self.disk._vdevType == vmdisks.Disk.DEVICE_CDROM:
            widget.set_sensitive(False)
        else:
            widget.set_sensitive(not exists)
        widget.set_value(gb)
        widget = self.glade.get_widget('disk-ro')
        widget.set_active(self.disk.ro)
        widget.set_sensitive(self.disk._vdevType != vmdisks.Disk.DEVICE_CDROM)
        widget = self.glade.get_widget('disk-sparse')
        widget.set_active(not self.disk.args.has_key('sparse') or self.disk.args['sparse'])
        widget.set_sensitive(not exists and self.disk._vdevType != vmdisks.Disk.DEVICE_CDROM)

    def browse_existing_disk(self, ignore=None):
        self._browse_disk(mode=gtk.FILE_CHOOSER_ACTION_OPEN)
    def browse_new_disk(self, ignore=None):
        self._browse_disk(mode=gtk.FILE_CHOOSER_ACTION_SAVE)
    def _browse_disk(self, mode):
        file = self.glade.get_widget('disk-location').get_child().get_text()
        (file, folder) = chooser(vminstall.msg.locate_disk,
                                 self.disk_edit_widget.get_toplevel(),
                                 file=file, mode=mode, folder=self.disk_folder)
        if file != None:
            self.disk_folder = folder
            self.glade.get_widget('disk-location').get_child().set_text(file)

    def populate_opt_media(self, model):
        if not self.bus:
            try:
                block_devs = os.listdir('/sys/block')
            except:
                return
            for bd in block_devs:
                try:
                    bdf = open('/sys/block/%s/removable' % bd)
                except:
                    continue
                if bdf.readline().strip() == "1":
                    dev = '/dev/%s' % bd
                    if not self._default_cdrom:
                        self._default_cdrom = dev
                    model.append([dev, dev, True, dev])
            return

        vollabel = {}
        volpath = {}

        # Track device add/removes so we can detect newly inserted CD media
        self.hal_iface.connect_to_signal("DeviceAdded", self._device_added)
        self.hal_iface.connect_to_signal("DeviceRemoved", self._device_removed)

        # Find info about all current present media
        vminstall.log.debug('Querying all present media...')
        for d in self.hal_iface.FindDeviceByCapability("volume"):
            vminstall.log.debug('Getting a volume...')
            vol = self.bus.get_object("org.freedesktop.Hal", d)
            vminstall.log.debug('Got volume.')
            if vol.GetPropertyBoolean("volume.is_disc") and \
                   vol.GetPropertyBoolean("volume.disc.has_data"):
                vminstall.log.debug('a disc with data')
                devnode = vol.GetProperty("block.device")
                label = vol.GetProperty("volume.label")
                if label == None or len(label) == 0:
                    label = devnode
                vollabel[devnode] = label
                volpath[devnode] = d
            else:
                vminstall.log.debug('not a disc with data')

        vminstall.log.debug('Querying all CDROMs...')
        for d in self.hal_iface.FindDeviceByCapability("storage.cdrom"):
            dev = self.bus.get_object("org.freedesktop.Hal", d)
            devnode = dev.GetProperty("block.device")
            if vollabel.has_key(devnode):
                model.append([devnode, vollabel[devnode], True, volpath[devnode]])
            else:
                model.append([devnode, vminstall.msg.no_media, False, None])

    def _device_added(self, path):
        vol = self.bus.get_object("org.freedesktop.Hal", path)
        if vol.QueryCapability("volume"):
            if vol.GetPropertyBoolean("volume.is_disc") and \
                   vol.GetPropertyBoolean("volume.disc.has_data"):
                devnode = vol.GetProperty("block.device")
                label = vol.GetProperty("volume.label")
                if label == None or len(label) == 0:
                    label = devnode

                cdlist = self.glade.get_widget("cd-path")
                model = cdlist.get_model()

                # Search for the row with matching device node and
                # fill in info about inserted media
                for row in model:
                    if row[0] == devnode:
                        row[1] = label
                        row[2] = True
                        row[3] = path

    def _device_removed(self, path):
        vol = self.bus.get_object("org.freedesktop.Hal", path)
        cdlist = self.glade.get_widget("cd-path")
        model = cdlist.get_model()

        active = cdlist.get_active()
        idx = 0
        # Search for the row containing matching HAL volume path
        # and update (clear) it, de-activating it if its currently
        # selected
        for row in model:
            if row[3] == path:
                row[1] = vminstall.msg.no_media
                row[2] = False
                row[3] = None
                if idx == active:
                    cdlist.set_active(-1)
            idx = idx + 1

