# GObject wrapper of installation job
#
# Copyright 2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gobject
import threading
import vminstall.log

class GJob(gobject.GObject):
    """This class starts a new thread for the VM creation job.  Events
    during the creation are emitted as GTK signals.
    """
    
    __gsignals__ = {
        'install-progress-change': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, (str, float)),
        'install-progress-close': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ()),
        'install-show-console': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, (str,)),
        'install-show-error': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ()),
        'install-done': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
    }

    def __init__(self, job, defaults, errors):
        self.__gobject_init__()
        self.job = job
        self.defaults = defaults
        self.errors = errors

    def start(self):
        def _getSettingsCB(job):
            return self.defaults
        def _errorCB(job, text):
            self.errors.insert(0, text)
            self.emit('install-show-error')
        def _stateChangeCB(job, state):
            self.emit('install-progress-change', state[2], -1.0)
        def _progressChangeCB(job, fraction):
            self.emit('install-progress-change', '', fraction)
        def _consoleCB(job):
            self.emit('install-show-console', job.guest.uuid)
        def _finish_worker():
            try:
                try:
                    self.job.run(GetSettingsCB=_getSettingsCB,
                                 StateChangeCB=_stateChangeCB,
                                 ConsoleCB=_consoleCB,
                                 ErrorCB=_errorCB,
                                 ProgressCB=_progressChangeCB)
                except Exception, e:
                    vminstall.log.error("VM creation job failed: %s" % str(e))
                    self.emit('install-progress-close')
                    self.errors.insert(0, e)
                    self.emit('install-show-error')
                else:
                    vminstall.log.info("VM creation job finished.")
                    self.emit('install-progress-close')
            finally:
                self.emit('install-done')
        thread = threading.Thread(target=_finish_worker)
        thread.setDaemon(True)
        thread.start()
