# GTK-based interface for vm-install
#
# Initially derived from virt-manager, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk
import gobject
import gtk.gdk
import gtk.glade
import os
import string
import sys
import threading

from disk_widgets import *
from gjob import GJob
from nic_widgets import *
from vminstall.job import Job
from vminstall.options import *
from vminstall.VMDefaults import VMDefaults as VMDefaults
import vminstall
import vminstall.console
import vminstall.nics
import vminstall.options
import vminstall.xend


def strify(name):
    if not name:
        return ''
    return str(name)

def set_combo_from_list(combo, items):
    """Setup a ComboBox or ComboBoxEntry based on a list of strings."""
    model = gtk.ListStore(gobject.TYPE_STRING)
    for item in items:
        model.append([item])
    combo.set_model(model)
    if type(combo) == gtk.ComboBoxEntry:
        combo.set_text_column(0)
    elif type(combo) == gtk.ComboBox:
        combo.clear()
        cell = gtk.CellRendererText()
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'text', 0)

#class NotebookPage:
#    def __init__(self, key, title):
#        self.key = key
#        self.root = gtk.VBox(homogeneous=False)
#        align = gtk.Alignment(xalign=0.0, yalign=0.5, xscale=0.0, yscale=1.0)
#        align.set_padding(5, 5, 6, 6)
#        label = gtk.Label()
#        label.set_single_line_mode(True)
#        label.set_markup('<span weight="heavy" size="xx-large" foreground="#FFF" background="#000">%s</span>' % title)
#        label.set_padding(5, 6)
#        align.add(label)
#        event = gtk.EventBox()
#        event.add(align)
#        event.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('#000'))
#        self.root.pack_start(event, expand=False)
#        align = gtk.Alignment(xalign=0.0, yalign=0.0, xscale=1.0, yscale=1.0)
#        align.set_padding(0, 0, 20, 0)
#        self.content = align
#        self.root.pack_start(self.content, expand=True)
#        self.root.show_all()

class VMCreate(gobject.GObject):
    __gsignals__ = {
        'action-show-console': (gobject.SIGNAL_RUN_FIRST,
                                gobject.TYPE_NONE, (str,str)),
        'action-show-terminal': (gobject.SIGNAL_RUN_FIRST,
                                gobject.TYPE_NONE, (str,str)),
        'vmmcreate-closing': (gobject.SIGNAL_RUN_FIRST,
                              gobject.TYPE_NONE, (int,)),
    }

    def refresh_xen_info(self):
        try:
            self.xen_info = vminstall.xend.connection.info()
        except:
            # An error will be displayed later, once the GUI is up.
            pass

    def __init__(self, job=None, options=None, defaults=None, virtman=False, key=0):
        self.__gobject_init__()

        self.job = job or Job(background=False)
        if options is None:
            options = vminstall.options.Options()
            defaults = vminstall.options.Options()
            vminstall.options.calculate(options, defaults)
        self.options = options
        self.defaults = defaults
        self.original_os_type = self.options.os_type

        self.xen_info = None
        self.refresh_xen_info()

        self.virtman = virtman
        self.key = key

        self.current_disk_index = None
        self.disk_caller_page = None
        self.nics_view = None
        self.current_nic = None
        self.current_nic_index = None

        self.os_settings_folder = os.getcwd()

        self.WINDOW_NAME = 'vmm-create'
        self.window = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', self.WINDOW_NAME)
        self.topwin = self.window.get_widget(self.WINDOW_NAME)
        self.topwin.set_title(vminstall.msg.title_intro)

        self.summary = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', 'summary-vbox')
        self.summary_vbox = self.summary.get_widget('summary-vbox')

        self.disks_widget = DisksWidget(allow_reorder=True, remove_fn=self.do_remove_disk,
                                        edit_fn=self.do_edit_disk, add_cdrom_fn=self.do_new_cdrom,
                                        add_harddisk_fn=self.do_new_harddisk)
        self.window.get_widget('disks-vbox').pack_end(self.disks_widget.get_widget())
        self.disk_edit_widget = DiskEditWidget()
        self.window.get_widget('disk-edit-vbox').pack_end(self.disk_edit_widget.get_widget())

        self.summary.signal_autoconnect({
            'on_headline_method_button_release_event': self.do_method,
            'on_headline_vm_properties_button_press_event': self.do_vm_properties,
            'on_headline_hardware_button_press_event': self.do_hardware,
            'on_headline_graphics_button_release_event': self.do_graphics,
            'on_headline_disks_button_press_event': self.do_disks,
            'on_headline_network_button_press_event': self.do_network,
            'on_headline_os_button_press_event': self.do_os,
            })

        self.window.signal_autoconnect({
            'on_create_pages_switch_page': self.page_changed,
            'on_create_cancel_clicked': self.cancel,
            'on_vmm_create_delete_event': self.close,
            'on_create_back_clicked': self.backward,
            'on_create_forward_clicked': self.forward,
            'on_create_apply_clicked': self.apply,
            'on_create_ok_clicked': self.okay,
            'on_pv_src_disk_add_clicked': self.do_new_pv_cdrom,
            'on_fv_src_disk_add_clicked': self.do_new_fv_cdrom,
            'on_button_nic_up_clicked': self.do_nic_up,
            'on_button_nic_down_clicked': self.do_nic_down,
            'on_button_new_nic_clicked': self.do_new_nic,
            'on_button_edit_nic_clicked': self.do_edit_nic,
            'on_button_delete_nic_clicked': self.do_delete_nic,
            'on_fv_install_source_focus_in': self.activate_fv_pxe_src,
            'on_pv_install_source_focus_in': self.activate_pv_network_src,
            'on_create_memory_max_value_changed': self.set_max_memory,
            'on_fv_media_toggled': self.change_fv_media_type,
            'on_pv_media_toggled': self.change_pv_media_type,
            'on_os_settings_browse_clicked': self.browse_os_settings,
            'on_os_type_list_button_release_event': self.select_os_type,
            })

        self.set_initial_state()

    def set_initial_state(self):
        self.notebook = self.window.get_widget('create-pages')
        self.notebook.set_show_tabs(False)
        self.pages = {}
        pages = self.notebook.get_n_pages()
        for i in xrange(pages):
            child = self.notebook.get_nth_page(i)
            label = self.notebook.get_tab_label_text(child)
            self.pages[label] = i

#        page = NotebookPage('test', "test")
#        self.notebook.insert_page(page.root, gtk.Label(page.key), 0)
#        page.content.add(gtk.Label("test"))
#        page.root.show_all()

        prompt = vminstall.msg.prompt
        bold_prompt = vminstall.msg.bold_prompt
        small = vminstall.msg.small

        # Set translated labels.  Not doing this within glade, so it's easier
        # to keep CLI & GUI in sync.
        black = gtk.gdk.color_parse('#000')
        for title in ('intro', 'install', 'os_type', 'summary', 'method',
                      'name', 'hardware', 'graphics', 'disks', 'disk',
                      'nics', 'nic_edit', 'fv_install', 'pv_install', 'pv_runtime'):
            widget = self.window.get_widget('title_%s' % title)
            widget.modify_bg(gtk.STATE_NORMAL, black)
            label = widget.get_child()
            label.set_markup('<span weight="heavy" size="xx-large" foreground="#FFF">%s</span>' % eval('vminstall.msg.title_%s' % title))
        # 'os' is set later, since it can be 'os_install' or 'os_runtime'
        for name in ('method', 'name', 'hardware', 'graphics', 'disks', 'nics'):
            title = self.summary.get_widget('summary-title-%s' % name)
            text = eval('vminstall.msg.title_%s' % name)
            self._set_summary_title(title, text)
        for name in ('memory', 'max_memory', 'vcpus', 'operating_system',
                     'installation_source', 'automated_installation', 'extra_args'):
            widget = self.summary.get_widget('%s_prompt' % name)
            text = eval('vminstall.msg.%s_prompt' % name)
            widget.set_markup(bold_prompt(text))

        def set_text(name, widget=None, text=None, modifier=None, markup=False):
            if widget is None:
                widget = self.window.get_widget(name)
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        def set_radio_text(name, text=None, modifier=None, markup=False):
            widget = self.window.get_widget(name).get_children()[0]
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        set_radio_text('graphics_hw_cirrus')
        set_radio_text('graphics_hw_none')
        set_radio_text('graphics_hw_para')
        set_radio_text('graphics_hw_vesa')
        set_radio_text('nic-mac-random', text=vminstall.msg.random_mac)
        set_radio_text('nic-mac-specified', text=vminstall.msg.specified_mac)
        set_radio_text('os_install_no')
        set_radio_text('os_install_yes')
        set_radio_text('pv-src-disk', text=vminstall.msg.virtual_disk, modifier=prompt)
        set_radio_text('pv-src-network', text=vminstall.msg.network_url, modifier=prompt)
        set_radio_text('fv-src-disk', text=vminstall.msg.virtual_disk, modifier=prompt)
        set_radio_text('fv-src-pxe', text=vminstall.msg.PXE_boot, modifier=prompt)
        set_radio_text('virt-method-fv', text=vminstall.msg.full_virtualization)
        set_radio_text('virt-method-pv', text=vminstall.msg.paravirtualization)
        set_text('available_cpus_prompt', modifier=prompt)
        set_text('available_memory_prompt', modifier=prompt)
        set_text('fv_install_help', text=vminstall.msg.install_media_help)
        set_text('graphics_hw_help')
        set_text('graphics_note_none', modifier=small, markup=True)
        set_text('graphics_note_para', modifier=small, markup=True)
        set_text('hardware_help')
        set_text('hardware-memory-title', text=vminstall.msg.memory, modifier=bold_prompt, markup=True)
        set_text('hardware-processors-note', text=vminstall.msg.processors_note, modifier=small, markup=True)
        set_text('hardware-processors-title', text=vminstall.msg.processors, modifier=bold_prompt, markup=True)
        set_text('i_extra_args_help', text=vminstall.msg.extra_args_help)
        set_text('i_extra_args_label', text=vminstall.msg.extra_args_prompt, modifier=prompt)
        set_text('install_help')
        set_text('intro_help1')
        set_text('intro_help2', markup=True)
        set_text('intro_help3', markup=True)
        set_text('intro_help4', markup=True)
        set_text('intro_help5', markup=True)
        set_text('intro_help6', text=vminstall.msg.intro_help6 % {'url': '<span underline="low">http://www.novell.com/documentation/technology/vm_server</span>'}, markup=True)
        set_text('max_memory_prompt', modifier=prompt)
        set_text('memory_prompt', modifier=prompt)
        set_text('method_help')
        set_text('method_no_full_bios', text=vminstall.msg.exc_arch_novt_bios, modifier=small, markup=True)
        set_text('method_no_full', text=vminstall.msg.exc_arch_novt, modifier=small, markup=True)
        set_text('method_no_full_os', text=vminstall.msg.exc_param_nofull, modifier=small, markup=True)
        set_text('method_no_para_os', text=vminstall.msg.exc_param_nopara, modifier=small, markup=True)
        set_text('name_help')
        set_text('name_name', text=vminstall.msg.name, modifier=prompt)
        set_text('nic_edit_help')
        set_text('nic-mac-prompt', text=vminstall.msg.mac_address, modifier=prompt)
        set_text('nic-name-prompt', text=vminstall.msg.name, modifier=prompt)
        set_text('nic-source-prompt', text=vminstall.msg.source, modifier=prompt)
        set_text('nic-type-prompt', text=vminstall.msg.type, modifier=prompt)
        set_text('os_settings_help')
        set_text('os_settings_label', text=vminstall.msg.os_settings_prompt)
        set_text('os_type_help')
        set_text('pv_install_help', text=vminstall.msg.install_media_help)
        set_text('r_extra_args_help', text=vminstall.msg.extra_args_help)
        set_text('r_extra_args_label', text=vminstall.msg.extra_args_prompt, modifier=prompt)
        set_text('summary_help', markup=True)
        set_text('vcpus_prompt', modifier=prompt)

        # If we don't have full-virt support disable the choice, and
        # display a message telling the user why it is not working
        if vminstall.caps.is_hvm_capable():
            self.window.get_widget("virt-method-fv").set_sensitive(True)
            self.window.get_widget("virt-method-fv-unsupported").hide()
            self.window.get_widget("virt-method-fv-disabled").hide()
        else:
            self.window.get_widget("virt-method-fv").set_sensitive(False)
            flags = vminstall.caps.cpu_flags()
            if "vmx" in flags or "svm" in flags:
                self.window.get_widget("virt-method-fv-unsupported").hide()
                self.window.get_widget("virt-method-fv-disabled").show()
            else:
                self.window.get_widget("virt-method-fv-unsupported").show()
                self.window.get_widget("virt-method-fv-disabled").hide()

        # Add OSs to os_type notebook page
        default = None
        groups = {}
        model = gtk.TreeStore(str, str)
        for os_type in VMDefaults.keys():
            group_name = eval('vminstall.msg.os_%s' % VMDefaults[os_type].group())
            if not groups.has_key(group_name):
                groups[group_name] = []
            groups[group_name].append(os_type)
        group_names = vminstall.util.sort_numeric_aware(groups.keys())
        for group_name in group_names:
            piter = model.append(None, [group_name, ''])
            oss = map(lambda os: (VMDefaults[os].name(), os), groups[group_name])
            oss = vminstall.util.sort_tuples_by_nth(oss, 0)
            for os in oss:
                model.append(piter, os)
                if self.defaults.os_type == os[1]:
                    default = (group_names.index(group_name), oss.index(os))
        treeview = self.window.get_widget('os-type-list')
        treeview.set_model(model)
        column = gtk.TreeViewColumn('os name', gtk.CellRendererText(), text=0)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        treeview.append_column(column)
        treeview.set_search_column(0)
        if default is not None:
            treeview.collapse_all()
            treeview.expand_row(default[:-1], False)
            treeselection = treeview.get_selection()
            treeselection.select_path(default)
        treeview.show_all()

        # Add NICs summary
        self.nics_view = NicsView([])
        self.window.get_widget('nics-scrolled-window').get_child().add(self.nics_view.view)
        self.nics_view.view.show_all()

        # Add summary
        scroll = self.window.get_widget('scrolled-summary')
        scroll.add_with_viewport(self.summary_vbox)

        self._reset_state()

    def _reset_state(self):
        self.notebook.set_current_page(self.pages['intro'])
        # Notebook is not yet displayed, so signal doesn't get emitted, so
        # load page defaults manually.
        self.load_page_defaults(self.pages['intro'])
        self.change_pv_media_type()

    def reset_state(self):
        self._reset_state()
        self.options = vminstall.options.Options()
        self.defaults = vminstall.options.Options()

    def show(self):
        self.topwin.show()

        hand = gtk.gdk.Cursor(gtk.gdk.HAND2)
        for name in ('method', 'name', 'hardware', 'graphics', 'disks', 'nics', 'os'):
            widget = self.summary.get_widget('summary-title-%s-eb' % name)
            widget.realize()
            widget.window.set_cursor(hand)

        self.topwin.present()

    def is_visible(self):
        if self.topwin.flags() & gtk.VISIBLE:
           return True
        return False

    def _set_summary_title(self, widget, text):
        widget.set_markup('<big><span foreground="blue">%s</span></big>' % text)

    def graphics_str(self):
        g = self.defaults.graphics
        text = eval('vminstall.msg.graphics_hw_' + g)
        return text

    def disks_str(self):
        text = ''
        if self.defaults.disks:
            for i in xrange(len(self.defaults.disks)):
                text += '%d: %s\n' % (i+1, self.defaults.disks[i])
        if not text:
            text = vminstall.msg.none
        return text.strip()

    def nics_str(self):
        text = ''
        if self.defaults.nics:
            for i in xrange(len(self.defaults.nics)):
                text += '%d: %s\n' % (i+1, str(self.defaults.nics[i]))
        if not text:
            text = vminstall.msg.none
        return text.strip()

    def select_os_type(self, widget, event):
        if event.type == gtk.gdk._2BUTTON_PRESS:
            self.forward()

    def do_new_pv_cdrom(self, button):
        self.disk_caller_page = 'pv_install'
        self.disks_widget.set_disks(self.defaults.disks)
        self.do_new_cdrom(button)

    def do_new_fv_cdrom(self, button):
        self.disk_caller_page = 'fv_install'
        self.disks_widget.set_disks(self.defaults.disks)
        self.do_new_cdrom(button)

    def do_new_cdrom(self, button):
        disk = vmdisks.Disk(pdev=self.disk_edit_widget.default_cdrom(),
                            vdevType=vmdisks.Disk.DEVICE_CDROM,
                            full_virt=self.defaults.full_virt,
                            disk_group=self.disks_widget.get_disks())
        self._do_new_disk(disk)

    def do_new_harddisk(self, button):
        disk = vmdisks.Disk(pdev=None, vdevType=vmdisks.Disk.DEVICE_DISK,
                            full_virt=self.defaults.full_virt,
                            disk_group=self.disks_widget.get_disks())
        self._do_new_disk(disk)

    def _do_new_disk(self, disk):
        disk.update(vmname=self.defaults.name)
        self.disk_edit_widget.set(disk)
        self.notebook.set_current_page(self.pages['disk'])
        return True

    def do_remove_disk(self, disk):
        remaining_disks = self.disks_widget.get_disks()
        for disk in remaining_disks:
            disk.disk_group = remaining_disks

    def do_edit_disk(self, button):
        (disk, index) = self.disks_widget.get_selected()
        if disk is None:
            return False
        self.current_disk_index = index
        self.disk_edit_widget.set(disk.copy())
        log.debug("Editing disk '%s'" % str(disk))
        self.notebook.set_current_page(self.pages['disk'])
        return True

    def load_page_defaults(self, page_number):
        """Initializes a page, loading the settings from 'defaults'."""
        self.set_page_buttons(page_number)
        if page_number == self.pages['install']:
            if self.defaults.install:
                self.window.get_widget('os_install_yes').set_active(True)
            else:
                self.window.get_widget('os_install_no').set_active(True)
        elif page_number == self.pages['os_type']:
            treeview = self.window.get_widget('os-type-list')
            treeselection = treeview.get_selection()
            rows = treeselection.get_selected_rows()[1]
            if len(rows):
                rows = rows[0]
                treeview.scroll_to_cell(rows)
        elif page_number == self.pages['summary']:
            if self.defaults.full_virt:
                text = vminstall.msg.fully_virtualized
            else:
                text = vminstall.msg.paravirtualized
            self.summary.get_widget("summary-method").set_text(text)
            self.summary.get_widget("summary-name").set_text(strify(self.defaults.name))
            self.summary.get_widget("summary-initial-memory").set_text(str(self.defaults.memoryMB) + " MB")
            self.summary.get_widget("summary-max-memory").set_text(str(self.defaults.max_memoryMB) + " MB")
            self.summary.get_widget("summary-virtual-cpus").set_text(str(int(self.defaults.vcpus)))
            self.summary.get_widget("summary-graphics").set_text(self.graphics_str())
            self.summary.get_widget("summary-disks").set_text(self.disks_str())
            self.summary.get_widget("summary-nics").set_text(self.nics_str())

            title = self.summary.get_widget('summary-title-os')
            self.summary.get_widget("summary-os-type").set_text(VMDefaults[self.defaults.os_type].name())
            src_row = ('installation_source_prompt', 'summary-install-source')
            auto_row = ('automated_installation_prompt', 'summary-os-settings')
            args_row = ('extra_args_prompt', 'summary-extra-args')
            def show_row(row):
                for name in row:
                    self.summary.get_widget(name).show()
            def hide_row(row):
                for name in row:
                    self.summary.get_widget(name).hide()
            if self.defaults.install:
                self._set_summary_title(title, vminstall.msg.title_os_install)
                src = strify(self.defaults.source)
                if self.defaults.pxe_boot:
                    src = vminstall.msg.PXE_boot
                self.summary.get_widget("summary-install-source").set_text(src)
                show_row(src_row)
                if self.defaults.full_virt:
                    hide_row(auto_row)
                else:
                    text = ''
                    if self.defaults.os_settings:
                        auto_inst_type = VMDefaults[self.defaults.os_type].auto_install()
                        if auto_inst_type is None:
                            text = vminstall.msg.exc_param_noauto
                        else:
                            text = '%s, using %s' % (auto_inst_type, self.defaults.os_settings)
                    self.summary.get_widget("summary-os-settings").set_text(text)
                    show_row(auto_row)
            else:
                self._set_summary_title(title, vminstall.msg.title_os_runtime)
                hide_row(src_row)
                hide_row(auto_row)
            if self.defaults.full_virt:
                hide_row(args_row)
            else:
                self.summary.get_widget("summary-extra-args").set_text(self.defaults.extra_args)
                show_row(args_row)
        elif page_number == self.pages['method']:
            for method in (('virt-method-pv-unsupported-by-os', 'virt-method-pv', self.defaults.para_possible),
                           ('virt-method-fv-unsupported-by-os', 'virt-method-fv', self.defaults.full_possible)):
                warn_widget = self.window.get_widget(method[0])
                widget = self.window.get_widget(method[1])
                if method[2]():
                    widget.set_sensitive(True)
                    warn_widget.hide()
                else:
                    widget.set_sensitive(False)
                    warn_widget.show()
            if self.defaults.full_virt:
                name = 'virt-method-fv'
            else:
                name = 'virt-method-pv'
            self.window.get_widget(name).set_active(True)
        elif page_number == self.pages['name']:
            w = self.window.get_widget('create-vm-name')
            w.connect('activate', self.apply)
            w.set_text(strify(self.defaults.name))
            w.grab_focus()
        elif page_number == self.pages['hardware']:
            self.refresh_xen_info()
            self.window.get_widget("create-cpus-physical").set_text(str(self.xen_info['nr_cpus']))
            if self.defaults.full_virt:
                avail_mb = self.xen_info['max_hvm_memory']
            else:
                avail_mb = self.xen_info['max_para_memory']
            self.window.get_widget("host-memory-free").set_text('%d MB' % avail_mb)
            self.window.get_widget("memory-max").set_value(self.defaults.max_memoryMB)
            self.window.get_widget("memory-startup").set_value(self.defaults.memoryMB)
            self.window.get_widget("create-vcpus").set_value(self.defaults.vcpus)
            self.set_max_memory()
        elif page_number == self.pages['graphics']:
            fv = self.defaults.full_virt
            for name in ('graphics_hw_cirrus', 'graphics_hw_vesa'):
                self.window.get_widget(name).set_sensitive(fv)
            for name in ('graphics_hw_para', 'graphics_hw_para_help'):
                self.window.get_widget(name).set_sensitive(not fv)
            self.window.get_widget('graphics_hw_' + self.defaults.graphics).set_active(True)
        elif page_number == self.pages['disks']:
            disks = self.disks_widget.get_disks()
            for d in disks:
                d.disk_group = disks
            max_disks = len(vmdisks.get_possible_vdevs(self.defaults.full_virt))
            self.disks_widget.set_max_disks(max_disks)
        elif page_number == self.pages['disk']:
            self.disk_edit_widget.refresh()
        elif page_number == self.pages['nics']:
            self.window.get_widget('nics-scrolled-window').get_child().grab_focus()
            self.current_nic = self.current_nic_index = None
            num_nics = self.nics_view.rows()
            for button in ('button-nic-up', 'button-nic-down'):
                self.window.get_widget(button).set_sensitive(num_nics > 1)
            for button in ('button-edit-nic', 'button-delete-nic'):
                self.window.get_widget(button).set_sensitive(num_nics > 0)
            for button in ('button-new-nic', ):
                self.window.get_widget(button).set_sensitive(num_nics < vminstall.xend.max_nics)
        elif page_number == self.pages['nic_edit']:
            nic = self.current_nic
            assert(nic is not None)
            self.window.get_widget('nic-name').set_text(str(nic.number))

            combo = self.window.get_widget('nic-type')
            nic_models = vminstall.util.sort_dict_keys_by_value(vminstall.nics.XenNic.models)
            set_combo_from_list(combo, [vminstall.nics.XenNic.models[m] for m in nic_models])
            combo.set_active(nic_models.index(nic.model))

            bridges = [vminstall.msg.default]
            bridges.extend(vminstall.nics.get_bridges())
            combo = self.window.get_widget('nic-source')
            set_combo_from_list(combo, bridges)
            if nic.bridge in bridges:
                combo.set_active(bridges.index(nic.bridge))
            else:
                combo.set_active(0)

            if nic.mac is None:
                self.window.get_widget('nic-mac-random').set_active(True)
            else:
                self.window.get_widget('nic-mac-specified').set_active(True)
                bytes = nic.mac.split(':')
                assert(len(bytes) == 6)
                for i in (1, 2, 3):
                    self.window.get_widget('mac%d' % i).set_text(bytes[-i])
        elif page_number == self.pages['fv_install']:
            combo = self.window.get_widget('fv-src-disk-box')
            self.src_disks = vmdisks.get_fv_boot_disks(self.defaults.disks)
            src_desc = [str(disk) for disk in self.src_disks]
            set_combo_from_list(combo, src_desc)
            source = self.defaults.source
            if len(self.src_disks):
                if type(source) is vmdisks.Disk and source in self.src_disks:
                    combo.set_active(self.src_disks.index(source))
                else:
                    combo.set_active(0)
            if self.options.pxe_boot:
                self.window.get_widget('fv-src-pxe').set_active(True)
            else:
                self.window.get_widget('fv-src-disk').set_active(True)
        elif page_number == self.pages['pv_install']:
            combo = self.window.get_widget('pv-src-disk-box')
            self.src_disks = filter(lambda disk: disk.exists(), self.defaults.disks)
            src_desc = [str(disk) for disk in self.src_disks]
            set_combo_from_list(combo, src_desc)
            source = self.defaults.source
            if len(self.src_disks):
                if type(source) is vmdisks.Disk and source in self.src_disks:
                    combo.set_active(self.src_disks.index(source))
                else:
                    combo.set_active(0)
            if type(source) is vmdisks.Disk:
                self.window.get_widget('pv-src-disk').set_active(True)
            else:
                self.window.get_widget('pv-src-network').set_active(True)
                self.window.get_widget('pv-src-network-url').set_text(strify(source))
            autotype = VMDefaults[self.defaults.os_type].auto_install()
            sens = (autotype is not None)
            if not autotype:
                autotype = vminstall.msg.os_settings_prompt
            autotype += ':'
            self.window.get_widget('os_settings_label').set_text(autotype)
            self.window.get_widget('os-settings').set_text(self.defaults.os_settings)
            for l in ('os_settings_help', 'os-settings-hbox'):
                self.window.get_widget(l).set_sensitive(sens)
            self.window.get_widget('i-extra-args').set_text(self.defaults.extra_args)
        elif page_number == self.pages['pv_runtime']:
            w = self.window.get_widget('r-extra-args')
            w.connect('activate', self.apply)
            w.set_text(self.defaults.extra_args)
            w.grab_focus()

    ##
    ##  Buttons
    ##

    def cancel(self, ignore=None):
        page_number = self.notebook.get_current_page()
        if page_number in self.intro_pages():
            self.close()
        elif page_number == self.pages['disk']:
            disk_caller_page = self.disk_caller_page or 'disks'
            self.notebook.set_current_page(self.pages[disk_caller_page])
            self.current_disk_index = None
        elif page_number == self.pages['nic_edit']:
            self.notebook.set_current_page(self.pages['nics'])
        else:
            self.notebook.set_current_page(self.pages['summary'])

    def okay(self, ignore=None):
        page_number = self.notebook.get_current_page()
        if page_number == self.pages['summary']:
            self.finish()
        else:
            self.apply()

    def apply(self, ignore=None):
        if self.save_page(self.notebook.get_current_page()):
            self.notebook.set_current_page(self.pages['summary'])

    def intro_pages(self):
        return [self.pages[name] for name in ['intro', 'install', 'os_type', 'summary']]

    def valid_intro_pages(self):
        valid_pages = []
        if not self.options.os_settings or not self.original_os_type:
            valid_pages.append(self.pages['intro'])
        if not self.options.os_settings:
            valid_pages.append(self.pages['install'])
        if not self.original_os_type:
            valid_pages.append(self.pages['os_type'])
        valid_pages.append(self.pages['summary'])
        return valid_pages

    def valid_forward_pages(self, current_page):
        pages = self.intro_pages()
        forward_pages = pages[pages.index(current_page)+1:]
        valid_pages = self.valid_intro_pages()
        return filter(lambda page: page in valid_pages, forward_pages)

    def valid_backward_pages(self, current_page):
        pages = self.intro_pages()
        pages.reverse()
        backward_pages = pages[pages.index(current_page)+1:]
        valid_pages = self.valid_intro_pages()
        return filter(lambda page: page in valid_pages, backward_pages)

    def forward(self, ignore=None):
        if self.xen_info is None:
            log.error("The hypervisor is not running")
            self._error_box(vminstall.msg.error, vminstall.msg.exc_xen_none)
            self.close()
            return
        current_page = self.notebook.get_current_page()
        if self.save_page(current_page):
            self.notebook.set_current_page(self.valid_forward_pages(current_page)[0])

    def backward(self, ignore=None):
        current_page = self.notebook.get_current_page()
        self.notebook.set_current_page(self.valid_backward_pages(current_page)[0])

    def close(self, ignore1=None, ignore2=None):
        log.info("vm-install #%d is closing..." % self.key)
        self.job.do_cancel(forceful=False)
        self.topwin.hide()
        self.topwin.destroy()
        self.emit('vmmcreate-closing', self.key)

    def set_page_buttons(self, page_number):
        if page_number in self.intro_pages():
            self.window.get_widget('create-apply').hide()
            forward = (len(self.valid_forward_pages(page_number)) > 0)
            backward = (len(self.valid_backward_pages(page_number)) > 0)
            self.window.get_widget('create-back').show()
            self.window.get_widget('create-back').set_sensitive(backward)
            self.window.get_widget("create-cancel").show()
            if page_number == self.pages['summary']:
                self.window.get_widget('create-ok').show()
                self.window.get_widget('create-forward').hide()
            else:
                self.window.get_widget('create-ok').hide()
                self.window.get_widget('create-forward').show()
                self.window.get_widget('create-forward').set_sensitive(forward)
            return
        if page_number == self.pages['disk']:
            self.window.get_widget("create-forward").hide()
            self.window.get_widget("create-back").hide
            self.window.get_widget("create-apply").hide()
            self.window.get_widget("create-ok").show()
            self.window.get_widget("create-cancel").show()
        else:
            self.window.get_widget("create-forward").hide()
            self.window.get_widget("create-back").hide()
            self.window.get_widget("create-ok").hide()
            self.window.get_widget("create-cancel").show()
            self.window.get_widget("create-apply").show()

    def do_method(self, widget, event):
        self.notebook.set_current_page(self.pages['method'])

    def do_vm_properties(self, widget, event):
        self.notebook.set_current_page(self.pages['name'])

    def do_hardware(self, widget, event):
        self.notebook.set_current_page(self.pages['hardware'])

    def do_graphics(self, widget, event):
        self.notebook.set_current_page(self.pages['graphics'])

    def do_disks(self, widget, event):
        self.disks_widget.set_disks(self.defaults.disks)
        self.notebook.set_current_page(self.pages['disks'])

    def do_nic_up(self, button):
        for row in self.nics_view.move_selected_up():
            nic = self.nics_view.get_row(row)
            self.nics_view.update(row, nic)
        return True

    def do_nic_down(self, button):
        for row in self.nics_view.move_selected_down():
            nic = self.nics_view.get_row(row)
            self.nics_view.update(row, nic)
        return True

    def do_new_nic(self, button):
        self.current_nic = vminstall.nics.XenNic(number=self.nics_view.rows()+1,
                                                 full_virt=self.defaults.full_virt)
        self.notebook.set_current_page(self.pages['nic_edit'])
        return True

    def do_edit_nic(self, button):
        (nic, index) = self.nics_view.get_selected()
        if nic is None:
            return False
        self.current_nic = nic.copy()
        self.current_nic_index = index
        log.debug("Editing nic '%s'" % str(self.current_nic))
        self.notebook.set_current_page(self.pages['nic_edit'])
        return True

    def do_delete_nic(self, button):
        self.nics_view.delete_selected()
        self.load_page_defaults(self.pages['nics'])
        return True

    def do_network(self, widget, event):
        nics = self.defaults.nics
        log.debug("Re-loading nics table with '%s'" % map(str, nics))
        self.nics_view.set(nics)
        self.notebook.set_current_page(self.pages['nics'])

    def do_os(self, widget, event):
        if self.defaults.install:
            if self.defaults.full_virt:
                self.notebook.set_current_page(self.pages['fv_install'])
            else:
                self.notebook.set_current_page(self.pages['pv_install'])
        else:
            if self.defaults.full_virt:
                pass  # No FV runtime configuration possible
            else:
                self.notebook.set_current_page(self.pages['pv_runtime'])

    def page_changed(self, notebook, page, page_number):
        self.load_page_defaults(page_number)

    def finish(self, ignore=None):
        try:
            vminstall.options.finalize(self.defaults)
        except Exception, e:
            self._exc_box(e)
            return

        # Create progress window
        def async_cancel(button):
            self.job.do_cancel(forceful=True)
        pbar_glade = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', "vmm-progress")
        pbar_glade.signal_autoconnect({
            'on_progress_cancel_clicked': async_cancel,
        })
        pbar_glade.get_widget('please_wait').set_text(vminstall.msg.please_wait)
        pbar_win = pbar_glade.get_widget("vmm-progress")
        pbar_win.connect('delete-event', lambda w,e: True)
        pbar_win.set_title(vminstall.msg.progress_title)

        # Show progress window
        self.topwin.set_sensitive(False)
        pbar_win.set_transient_for(self.topwin)
        pbar_win.present()
        pbar_win.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))

        # Create GObject Job wrapper
        pbar = pbar_glade.get_widget("pbar")
        label = pbar_glade.get_widget("status")
        errors = []

        # GTK 2.11 or PyGTK 2.10 seem to suffer from a threading bug.  Some
        # windowing functions are unsafe to call from another thread, even
        # while holding the GTK lock.  Use idle callbacks instead.  Novell bz
        # #306964
        def destroy_window(w):
            gtk.gdk.threads_enter()
            w.destroy()
            gtk.gdk.threads_leave()
        def hide_window(w):
            gtk.gdk.threads_enter()
            w.hide()
            gtk.gdk.threads_leave()
        def do_progress_change(src, text, fraction):
            gtk.gdk.threads_enter()
            log.debug("do_progress_change: %0.2f %s" % (fraction, text))
            if text:
                label.set_text(text)
            if fraction >= 0.0 and fraction <= 1.0:
                pbar.set_fraction(fraction)
            gtk.gdk.threads_leave()
        def do_progress_close(src):
            log.debug("do_progress_close")
            if pbar_win is None:
                return
            gtk.gdk.threads_enter()
            gobject.idle_add(destroy_window, pbar_win)
            self.topwin.set_sensitive(True)
            try:
                self.topwin.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.TOP_LEFT_ARROW))
            except:
                pass  # Parent's window may have closed
            gtk.gdk.threads_leave()
        def do_show_console(src, uuid):
            log.debug("do_show_console")
            if self.virtman:
                domid = vminstall.console.get_domid_from_uuid(uuid)
                signal = 'action-show-terminal'
                if (self.defaults.graphics != 'none' and \
                    vminstall.console.get_vnc_port_from_domid(domid, timeout=10.0) != None):
                    signal = 'action-show-console'
                log.debug("%s of %s via virt-manager" % (signal, uuid))
                def poke_virt_man():
                    self.emit(signal, 'xen', uuid)
                    gobject.idle_add(hide_window, pbar_win)
                    gobject.idle_add(hide_window, self.topwin)
                    return False  # Do not call me again
                # virt-manager (and libvirt) poll.  Asking virt-manager to open
                # the console too soon after the VM has started will fail.
                gobject.timeout_add(1100, poke_virt_man)
                if self.job.state[0] == Job.INSTALL:
                    wait_on_me = uuid
                else:
                    wait_on_me = None
            else:
                wait_on_me = self.job._show_console()
            if type(wait_on_me) is str:
                vminstall.console.wait_for_uuid_to_exit(wait_on_me)
            elif type(wait_on_me) is int:
                gobject.idle_add(hide_window, pbar_win)
                gobject.idle_add(hide_window, self.topwin)
                (pid, status) = os.waitpid(wait_on_me, 0)
        def _do_show_error(sem):
            gtk.gdk.threads_enter()
            log.debug("_do_show_error")
            e = errors.pop()
            log.debug("_do_show_error: error was '%s'" % str(e))
            if type(e) is str:
                self._error_box(vminstall.msg.error, e)
            else:
                self._exc_box(e)
            sem.release()
            gtk.gdk.threads_leave()
        def do_show_error(src):
            log.debug("do_show_error")
            sem = threading.Semaphore()
            sem.acquire()
            gobject.idle_add(_do_show_error, sem)
            sem.acquire()
        def _do_done():
            gtk.gdk.threads_enter()
            log.debug("_do_done: state was '%s'" % self.job.state[2])
            if self.job.state[0] == self.job.FINISHED:
                self.close()
            else:
                self.topwin.present()
            gtk.gdk.threads_leave()
        def do_done(src):
            log.debug("do_done")
            gobject.idle_add(_do_done)
        log.info("Starting VM creation job (key=%d)..." % self.key)
        gjob = GJob(self.job, self.defaults, errors)
        gjob.connect('install-progress-change', do_progress_change)
        gjob.connect('install-progress-close', do_progress_close)
        gjob.connect('install-show-console', do_show_console)
        gjob.connect('install-show-error', do_show_error)
        gjob.connect('install-done', do_done)
        gjob.start()

    def activate_fv_pxe_src(self, ignore1=None, ignore2=None):
        self.window.get_widget('fv-src-pxe').activate()

    def activate_pv_network_src(self, ignore1=None, ignore2=None):
        self.window.get_widget('pv-src-network').activate()

    def browse_os_settings(self, src, ignore=None):
        file = self.window.get_widget('os-settings').get_text()
        (file, folder) = chooser(vminstall.msg.locate_os_settings,
                                 self.topwin, file=file,
                                 mode=gtk.FILE_CHOOSER_ACTION_OPEN,
                                 folder=self.os_settings_folder)
        if file != None:
            self.os_settings_folder = folder
            self.window.get_widget('os-settings').set_text(file)

    def change_fv_media_type(self, ignore=None):
        if self.window.get_widget("fv-src-disk").get_active():
            self.window.get_widget("fv-src-disk-box").set_sensitive(True)
            self.window.get_widget("fv-src-disk-add").set_sensitive(True)
        elif self.window.get_widget("fv-src-pxe").get_active():
            self.window.get_widget("fv-src-disk-box").set_sensitive(False)
            self.window.get_widget("fv-src-disk-add").set_sensitive(False)

    def change_pv_media_type(self, ignore=None):
        if self.window.get_widget("pv-src-disk").get_active():
            self.window.get_widget("pv-src-disk-box").set_sensitive(True)
            self.window.get_widget("pv-src-disk-add").set_sensitive(True)
            self.window.get_widget("pv-src-network-url").set_sensitive(False)
        elif self.window.get_widget("pv-src-network").get_active():
            self.window.get_widget("pv-src-disk-box").set_sensitive(False)
            self.window.get_widget("pv-src-disk-add").set_sensitive(False)
            self.window.get_widget("pv-src-network-url").set_sensitive(True)

    def set_max_memory(self, src=None):
        if self.defaults.full_virt:
            avail_mb = self.xen_info['max_hvm_memory']
        else:
            avail_mb = self.xen_info['max_para_memory']
        maximum = self.window.get_widget("memory-max").get_adjustment()
        startup = self.window.get_widget("memory-startup").get_adjustment()
        if startup.value > maximum.value:
            startup.value = maximum.value
        if startup.value > avail_mb:
            startup.value = avail_mb
        startup.upper = min(maximum.value, avail_mb)

    def save_page(self, page_num):
        """Returns True if ok to proceed to next page, else False.
        Transactional w.r.t options/defaults.
        """
        ok = False
        old_options = self.options.copy()
        try:
            ok = self._save_page(page_num)
        except Exception, e:
            self._exc_box(e)
            self.options = old_options
        if not ok:
            return False

        old_defaults = self.defaults.copy()
        try:
            vminstall.options.calculate(self.options, self.defaults)
        except Exception, e:
            self._exc_box(e)
            self.options = old_options
            self.defaults = old_defaults
            return False
        return True

    def _save_page(self, page_num):
        """Saves data from the specified page into 'options'.
        On sucess, returns True.
        Not transactional on failure, w.r.t. options/defaults.
        """
        if page_num == self.pages['intro']:
            pass
        elif page_num == self.pages['install']:
            self.options.install = self.window.get_widget('os_install_yes').get_active()
        elif page_num == self.pages['os_type']:
            treeview = self.window.get_widget('os-type-list')
            selection = treeview.get_selection()
            (model, iter) = selection.get_selected()
            if iter is None:
                return False
            os_type = model.get_value(iter, 1)
            if not os_type:
                return False
            self.options.os_type = os_type
        elif page_num == self.pages['summary']:
            pass
        elif page_num == self.pages['hardware']:
            def get_spin_value(widget):
                widget.update()
                return widget.get_value()
            self.options.memoryMB = get_spin_value(self.window.get_widget("memory-startup"))
            self.options.max_memoryMB = get_spin_value(self.window.get_widget("memory-max"))
            self.options.vcpus = get_spin_value(self.window.get_widget("create-vcpus"))
        elif page_num == self.pages['graphics']:
            for hw in ('cirrus', 'none', 'para', 'vesa'):
                if self.window.get_widget('graphics_hw_' + hw).get_active():
                    self.options.graphics = hw
                    break
        elif page_num == self.pages['method']:
            full_virt = not self.window.get_widget('virt-method-pv').get_active()
            self.options.full_virt = full_virt
            if not self.options.full_virt:
                self.options.pxe_boot = False
        elif page_num == self.pages['name']:
            name = self.window.get_widget('create-vm-name').get_text()
            self.options.name = name
        elif page_num == self.pages['disks']:
            self.options.disks = self.disks_widget.get_disks()
            for d in self.options.disks:
                d.disk_group = self.options.disks
        elif page_num == self.pages['disk']:
            disk = self.disk_edit_widget.get()

            # Verify that the new disk does not conflict.  If it does, the
            # transaction will roll back, and GUI disk list hasn't yet been changed.
            self.options.disks = self.disks_widget.get_disks()
            if self.current_disk_index is not None:
                self.options.disks[self.current_disk_index] = disk
            else:
                self.options.disks.append(disk)
            vminstall.options.calculate(self.options, self.defaults)

            # It does not conflict, so safe to add it to the disks widget:
            if self.current_disk_index is not None:
                self.disks_widget.set_index(self.current_disk_index, disk)
                self.current_disk_index = None
            else:
                self.disks_widget.append(disk)

            if self.disk_caller_page:
                self.options.source = disk
                ret_page = self.disk_caller_page
                self.disk_caller_page = None
            else:
                ret_page = 'disks'
            vminstall.options.calculate(self.options, self.defaults)
            self.notebook.set_current_page(self.pages[ret_page])
            return None
        elif page_num == self.pages['nics']:
            self.options.nics = self.nics_view.get()
            # They've looked at the NICs, so don't allow the default to change if full_virt changes.
            for nic in self.options.nics:
                nic.set_model(nic.get_model())
            self.nics_view.set([])
            log.debug("Retrieved nic list from nic table: '%s'" % map(str, self.options.nics))
        elif page_num == self.pages['nic_edit']:
            nic = self.current_nic
            assert(nic is not None)
            nic_models = vminstall.util.sort_dict_keys_by_value(vminstall.nics.XenNic.models)
            nic.model = nic_models[self.window.get_widget('nic-type').get_active()]

            if self.window.get_widget('nic-mac-random').get_active():
                nic.mac = None
            else:
                mac1 = self.window.get_widget('mac1').get_text()
                mac2 = self.window.get_widget('mac2').get_text()
                mac3 = self.window.get_widget('mac3').get_text()
                nic.mac = '00:16:3e:%s:%s:%s' % (mac3, mac2, mac1)

            bridge = self.window.get_widget('nic-source').get_active()
            if bridge == 0:
                nic.bridge = None
            else:
                nic.bridge = vminstall.nics.get_bridges()[bridge-1]

            if self.current_nic_index is not None:
                self.nics_view.update(self.current_nic_index, self.current_nic)
            else:
                self.nics_view.append(self.current_nic)
            vminstall.options.calculate(self.options, self.defaults)
            self.notebook.set_current_page(self.pages['nics'])
            return None
        elif page_num == self.pages['fv_install']:
            if self.window.get_widget('fv-src-disk').get_active():
                i = self.window.get_widget('fv-src-disk-box').get_active()
                if i < 0:
                    raise vminstall.DiskResourceError(vminstall.DiskResourceError.E_NOT_FOUND)
                self.options.source = self.src_disks[i]
                self.options.pxe_boot = 0
            else:
                self.options.pxe_boot = 1
        elif page_num == self.pages['pv_install']:
            if self.window.get_widget('pv-src-disk').get_active():
                i = self.window.get_widget('pv-src-disk-box').get_active()
                if i < 0:
                    raise vminstall.DiskResourceError(vminstall.DiskResourceError.E_NOT_FOUND)
                src = self.src_disks[i]
            else:
                src = self.window.get_widget('pv-src-network-url').get_text()
            self.options.source = src
            self.options.os_settings = self.window.get_widget('os-settings').get_text()
            self.options.extra_args = self.window.get_widget('i-extra-args').get_text()
        elif page_num == self.pages['pv_runtime']:
            self.options.extra_args = self.window.get_widget('r-extra-args').get_text()

        return True

    def _exc_box(self, exc):
        if isinstance(exc, vminstall.Error):
            self._error_box(vminstall.msg.error, text=exc.title, details=exc.details)
        else:
            self._error_box(vminstall.msg.error, text=str(exc))

    def _error_box(self, title, text=None, details=None):
        self.topwin.present()
        message_box = gtk.MessageDialog(self.topwin,
                                        0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, title)
        if text:
            message_box.format_secondary_text(text)
        if details:
            expander = gtk.Expander("Details")
            tbuffer = gtk.TextBuffer()
            tbuffer.set_text(details)
            textview = gtk.TextView(tbuffer)
            expander.add(gtk.Label(details))
            expander.show_all()
            message_box.vbox.pack_end(expander)
        message_box.run()
        message_box.destroy()

def run(job, options, defaults):
    gtk.gdk.threads_init()
    assert(not options.background)
    window = VMCreate(job, options, defaults)
    window.topwin.connect("delete-event", gtk.main_quit)
    window.connect("vmmcreate-closing", gtk.main_quit)
    window.show()
    gtk.main()
