# Communicates with xend
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import time

try:
    # Xen 3.1+
    from xen.util.xmlrpcclient import ServerProxy
except:
    # Xen 3.0.4 and older
    from xen.util.xmlrpclib2 import ServerProxy

# Some Xen limits:
max_nics = 3

class Xend:
    """Thin abstraction to talk to Xend.

    DEPRECATED.  Migrate to libvirt instead.
    """
    XML_RPC_SOCKET = "/var/run/xend/xmlrpc.sock"

    def __init__(self):
        self.uri = 'httpu://%s' % Xend.XML_RPC_SOCKET
        self.server = ServerProxy(self.uri)

    def info(self):
        return dict(self.server.xend.node.info())

    def delete(self, domid):
        """Destroys a running instance; erases from xend"""
        try:
            self.server.xend.domain.delete(domid)
        except:
            pass

    def destroy(self, domid):
        """Destroys a running instance; does not erase from xend"""
        try:
            self.server.xend.domain.destroy(domid)
        except:
            pass

    def domain(self, domid):
        try:
            return self.server.xend.domain(domid, False)
        except:
            return None

    def domain_running(self, domid):
        domain = self.domain(domid)
        if not domain:
            return False
        for d in domain:
            if type(d) == list and len(d) > 1 and d[0] == 'state':
                # Any state means there is an instance...
                return len(d[1]) > 1
        return False

    def wait_domain(self, domid):
        #  This is ugly (polling instead of event driven) but XenAPI doesn't
        #  yet grok unix socket.
        #
        #  If domain and 'state' are both found, then assume VM is trying
        #  to come up, so keep waiting.  Otherwise, fail.
        for _ in xrange(10):
            domain = self.domain(domid)
            if not domain:
                return False
            wait = False
            for d in domain:
                if type(d) == list and len(d) > 1 and d[0] == 'state':
                    if len(d[1]) > 1 and not 'p' in d[1]:
                        return True
                    time.sleep(1.0)
                    wait = True
                    break
            if not wait:
                break
        return False

connection = Xend()
